{ AMHotKeys version 1.1
  Copyright  januari-march 1999 by Alexander Meeder - AM Software

  The idea for this component came from Arjen Broeze (e-mail: Arjen@Earthling.net) and I've used
  some of his code.

  With this component you can set one or more "HotKey's" for your application. The hotkey-events
  will be triggered even if another program has the inputfocus. There are two kind of hotkey-events:
  * One overall hotkey-event with a special parameter (the HotKey-index)
  * One special hotkey-event for every enabled hotkey combination

  You can also use this component to create HotKey-combinations at runtime (see the demo appl. that
  comes with this component)

  TIP: When you use this component, always test the key-combination, because some key-combinations
  have a special meaning within windows and won't work.

  TAMHotKey
  ---------
  property:
  * HotKeys (default property) is a THotKeys-collection, you can add new THotKeyItems at designtime
    using the collection editor
  * IniFileName is the name of the IniFile to use for storing the defined Hotkey-combinations
  * SaveLoadFrom is the method to use for saving the defined HotKey combinations

  event:
  * OnHotKeys is of type THotKeyEvent with one special Parameter: HotKeyIndex

  methods:
  * Load to load the previous save HotKey defenitions from IniFile of Register (see also the
    properties of TAMHotKey)
  * Save to save the currently defined HotKey defenitions to IniFile or Register (see also the
    properties of TAMHotKey)

  THotKeys (see property of TAMHotKey)
  --------
  property:
  * HotKeys is a collection of type THotKeyItems (see below for more info about THotKeyItem)

  method:
  * Add, adds a new THotKeyItem to the collection

  THotKeyItem
  -----------
  property:
  * Enabled (default false) is of type boolean and uses to en-/disable then hotkey
  * Modifiers (default none) is of type TModifiers and can be a combination of the Shift, Control, Alt and Win keys
  * VirtKey (default vkNone) is of type TVirtKey the actual hotkey (in combination with the modifier(s))

  event:
  * OnHotKey is of type TNotifyEvent and only triggered when there is code assigned to is and when
    the key-combination is Enabled

  if you have suggestions and/or problems please mail me at: ameeder@dds.nl

  Version history:
  1.1 Minor changes to make 'runtime-creation' and saving/loading of hotkeys available
  1.0 First release of this component
}

unit AMHotKey;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, IniFiles, AMReg;

type
  TAMHotKeys = class;

  TWMHotKey = record
    Msg: Cardinal;
    idHotKey: Word;
    Modifiers: Integer;
    VirtKey : Integer;
  end;

  TSaveLoad = (slIniFile, slRegister);
  TModifier = (modShift, modControl, modAlt, modWin);
  TModifiers = set of TModifier;

  TVirtKey = (vkNone, vkCancel, vkBack, vkTab, vkClear, vkReturn, vkPause, vkCapital, vkEscape,
              vkSpace, vkPrior, vkNext, vkEnd, vkHome, vkLeft, vkUp, vkRight, vkDown,
              vkSelect, vkExecute, vkSnapshot, vkInsert, vkDelete, vkHelp,
              vk0, vk1, vk2, vk3, vk4, vk5, vk6, vk7, vk8, vk9,
              vkA, vkB, vkC, vkD, vkE, vkF, vkG, vkH, vkI, vkJ, vkK, vkL, vkM,
              vkN, vkO, vkP, vkQ, vkR, vkS, vkT, vkU, vkV, vkW, vkX, vkY, vkZ,
              vkNumpad0, vkNumpad1, vkNumpad2, vkNumpad3, vkNumpad4,
              vkNumpad5, vkNumpad6, vkNumpad7, vkNumpad8, vkNumpad9,
              vkMultiply, vkAdd, vkSeparator, vkSubtract, vkDecimal, vkDivide,
              vkF1, vkF2, vkF3, vkF4, vkF5, vkF6, vkF7, vkF8, vkF9, vkF10, vkF11, vkF12,
              vkF13, vkF14, vkF15, vkF16, vkF17, vkF18, vkF19, vkF20, vkF21, vkF22, vkF23, vkF24,
              vkNumlock, vkScroll, vkApps);

  THotKeyEvent = procedure(Sender: TObject; HotKeyIndex: Word) of object;

  THotKeyItem = class(TCollectionItem)
  private
    FEnabled   : boolean;
    FModifiers : TModifiers;
    FRegistered: Boolean;
    FVirtKey   : TVirtKey;
    FHotKey: TNotifyEvent;

    procedure SetEnabled(Value: Boolean);
    procedure SetModifiers(Value: TModifiers);
    procedure SetVirtKey(Value: TVirtkey);

    procedure RegisterHK;
    procedure UnregisterHK;
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    procedure Assign(Source: TPersistent); override;
    procedure Free;

    property Registered: Boolean read FRegistered Default false;
  published
    property Enabled: boolean read FEnabled write SetEnabled Default false;
    property Modifiers: TModifiers read FModifiers write SetModifiers;
    property VirtKey: TVirtKey read FVirtKey write SetVirtKey default vkNone;
    property OnHotKey: TNotifyEvent read FHotKey write FHotKey;
  end;

  THotKeys = class(TCollection)
  private
    FAMHotKeys: TAMHotKeys;
    function GetItem(Index: Integer): THotKeyItem;
    procedure SetItem(Index: Integer; Value: THotKeyItem);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(AMHotKeys: TAMHotKeys);
    function Add: THotKeyItem;
    property HotKeys[Index: Integer]: THotKeyItem read GetItem write SetItem; default;
  end;

  TAMHotKeys = class(TComponent)
  private
    FHotKeys: THotKeys;
    FHotKeyEvents: THotKeyEvent;

    OldWndProc : TFarProc;
    NewWndProc : Pointer;
    FSaveLoad: TSaveLoad;
    FIniFileName: TFileName;
    FAMReg: TAMReg;

    procedure SetHotKeys(Value: THotKeys);
    procedure HookParent;
    procedure UnhookParent;
    procedure HookWndProc(var Message: TMessage);
    procedure SetSaveLoad(const Value: TSaveLoad);
  protected
    IniFile: TIniFile;
    Sects: TStrings;

    function GetKeyBasis: string;
    procedure LoadFromIniFile;
    procedure SaveToIniFile;
    procedure LoadFromRegister;
    procedure SaveToRegister;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    procedure Save;
    procedure Load;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property AMReg: TAMReg read FAMReg write FAMReg;
    property HotKeys: THotKeys read FHotKeys write SetHotKeys;
    property IniFileName: TFileName read FIniFileName write FIniFileName;
    property SaveLoadFrom: TSaveLoad read FSaveLoad write SetSaveLoad default slRegister;
    property OnHotKeys: THotKeyEvent read FHotKeyEvents write FHotKeyEvents;
  end;

var
  VirtKeyStr : array[TVirtKey] of string =
            ('None', 'Cancel', 'Back', 'Tab', 'Clear', 'Return', 'Pause', 'Capital', 'Escape',
             'Space', 'Prior', 'Next', 'End', 'Home', 'Left', 'Up', 'Right', 'Down',
             'Select', 'Execute', 'Snapshot', 'Insert', 'Delete', 'Help',
             '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
             'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M',
             'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z',
             'Numpad0', 'Numpad1', 'Numpad2', 'Numpad3', 'Numpad4',
             'Numpad5', 'Numpad6', 'Numpad7', 'Numpad8', 'Numpad9',
             'Multiply', 'Add', 'Separator', 'Subtract', 'Decimal', 'Divide',
             'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12',
             'F13', 'F14', 'F15', 'F16', 'F17', 'F18', 'F19', 'F20', 'F21', 'F22', 'F23', 'F24',
             'Numlock', 'Scroll', 'Apps');


procedure Register;

implementation

var
  VirtKeys : array[TVirtKey] of UInt =
             ($00, $03, $08, $09, $0C, $0D, $13, $14, $1B,
              $20, $21, $22, $23, $24, $25, $26, $27, $28,
              $29, $2B, $2C, $2D, $2E, $2F,
              $30, $31, $32, $33, $34, $35, $36, $37, $38, $39,
              $41, $42, $43, $44, $45, $46, $47, $48, $49, $4A,
              $4B, $4C, $4D, $4E, $4F, $50, $51, $52, $53, $54,
              $55, $56, $57, $58, $59, $5A,
              $60, $61, $62, $63, $64, $65, $66, $67, $68, $69,
              $6A, $6B, $6C, $6D, $6E, $6F,
              $70, $71, $72, $73, $74, $75, $76, $77, $78, $79, $7A, $7B,
              $7C, $7D, $7E, $7F, $80, $81, $82, $83, $84, $85, $86, $87,
              $90, $91, $5D);

// TAMHotKeys implementation
procedure TAMHotKeys.SetHotKeys(Value: THotKeys);
begin
  FHotKeys.Assign(Value);
end;

procedure TAMHotKeys.HookParent;
begin
  if not (csDesigning in ComponentState) then
  begin
    if (Owner as TForm) = nil then exit;

    OldWndProc := TFarProc(GetWindowLong((Owner as TWinControl).Handle, GWL_WNDPROC));
    NewWndProc := MakeObjectInstance(HookWndProc);
    SetWindowLong((Owner as TWinControl).Handle, GWL_WNDPROC, LongInt(NewWndProc));
  end;
end;

procedure TAMHotKeys.UnhookParent;
begin
  if ((Owner as TForm) <> nil) and Assigned(OldWndProc) then
     SetWindowLong((Owner as TWinControl).Handle, GWL_WNDPROC, LongInt(OldWndProc));
  if Assigned(NewWndProc) then
     FreeObjectInstance(NewWndProc);

  NewWndProc := nil;
  OldWndProc := nil;
end;

procedure TAMHotKeys.HookWndProc(var Message: TMessage);
var
  I: integer;
begin
  if (Owner as TWinControl) = nil then exit;

  with Message do
  begin
    // if the message is a WM_HOTKEY-message then its for us
    if Msg = WM_HOTKEY then
    begin
      with TWMHotKey(Message) do
      begin
        if Assigned(HotKeys[HotKeys.FindItemID(idHotKey).Index].FHotKey) then HotKeys[HotKeys.FindItemID(idHotKey).Index].FHotKey(Self);
        if Assigned(FHotKeyEvents) then FHotKeyEvents(Self, HotKeys.FindItemID(idHotKey).Index);
      end;
    end;

    // if the owner (a TForm-object) will be destroyed then we have to 'unregister' all
    // enabled HotKey-combinations by disabling them
    if Msg = WM_DESTROY then
      for I := 0 to HotKeys.Count - 1 do
        if HotKeys[I].Registered then HotKeys[I].Enabled := false;

    Result := CallWindowProc(OldWndProc, (Owner as TWinControl).Handle, Msg, wParam, lParam);
  end;
end;

procedure TAMHotKeys.SetSaveLoad(const Value: TSaveLoad);
begin
  FSaveLoad := Value;
end;

function TAMHotKeys.GetKeyBasis: string;
begin
  with FAMReg do
  begin
    Result := '\' + Group + '\' + Company + '\' + Application + '\';
    if User <> '' then Result := Result + User + '\';
    Result := Result + 'HotKeys\';
  end;
end;

procedure TAMHotKeys.Load;
begin
  if SaveLoadFrom = slIniFile then
    LoadFromIniFile
  else
    LoadFromRegister;
end;

procedure TAMHotKeys.LoadFromIniFile;
var
  HKItem : THotKeyItem;
  I: Integer;
begin
  if IniFileName <> '' then
  begin
    HotKeys.Clear;
    IniFile := TIniFile.Create(IniFileName);
    try
      with IniFile do
      begin
        Sects := TStringList.Create;
        try
          IniFile.ReadSections(Sects);

          for I := 0 to Sects.Count - 1 do
          begin
            HKItem := HotKeys.Add;
            with HKItem do
            begin
              VirtKey := TVirtKey(IniFile.ReadInteger(Sects[I], 'Virtkey', 0));
              if IniFile.ReadBool(Sects[I], 'Shift', False) then Modifiers := Modifiers + [modShift];
              if IniFile.ReadBool(Sects[I], 'Control', False) then Modifiers := Modifiers + [modControl];
              if IniFile.ReadBool(Sects[I], 'Alt', False) then Modifiers := Modifiers + [modAlt];
              if IniFile.ReadBool(Sects[I], 'Win', False) then Modifiers := Modifiers + [modWin];
              if IniFile.ReadBool(Sects[I], 'Enabled', False) then Enabled := True;
            end;
          end;
        finally
          Sects.Free;
        end;
      end;
    finally
      IniFile.Free;
    end;
  end
  else
    raise Exception.Create('IniFileName has no value!');
end;

procedure TAMHotKeys.LoadFromRegister;
var
  I: integer;
  HKItem: THotKeyItem;
  Keys: TStrings;
begin
  if Assigned(FAMReg) then
  begin
    HotKeys.Clear;
    with FAMReg do
    begin
      Active := True;
      Keys := TStringList.Create;
      try
        Reg.OpenKeyReadOnly(GetKeyBasis);
        Reg.GetKeyNames(Keys);
        for I := 0 to Keys.Count - 1 do
        begin
          HKItem := HotKeys.Add;
          Reg.OpenKeyReadOnly(GetKeyBasis + Keys[I]);
          with HKItem do
          begin
            VirtKey := TVirtKey(Reg.ReadInteger('VirtKey'));
            if Reg.ReadBool('Shift') then
              Modifiers := Modifiers + [modShift];
            if Reg.ReadBool('Control') then
              Modifiers := Modifiers + [modControl];
            if Reg.ReadBool('Alt') then
              Modifiers := Modifiers + [modAlt];
            if Reg.ReadBool('Win') then
              Modifiers := Modifiers + [modWin];
            Enabled := Reg.ReadBool('Enabled');
          end;
        end;
      finally
        Keys.Free;
        Active := False;
      end;
    end;
  end
  else
    raise Exception.Create('There is no AMReg-component assigned to AMReg-rpoperty!');
end;

procedure TAMHotKeys.Save;
begin
  if SaveLoadFrom = slIniFile then
    SaveToIniFile
  else
    SaveToRegister;
end;

procedure TAMHotKeys.SaveToIniFile;
var
  I: integer;
  Sect: string;
begin
  if IniFileName <> '' then
  begin
    IniFile := TIniFile.Create(IniFileName);
    try
      with IniFile do
      begin
        Sects := TStringList.Create;
        try
          IniFile.ReadSections(Sects);
          for I := 0 to Sects.Count - 1 do IniFile.EraseSection(Sects[I]);
        finally
          Sects.Free;
        end;

        for I := 0 to HotKeys.Count - 1 do
        begin
          with HotKeys[I] do
          begin
            Sect := DisplayName;
            WriteInteger(Sect, 'Virtkey', Ord(VirtKey));
            if (modShift in Modifiers) then
              WriteBool(Sect, 'Shift', True)
            else
              WriteBool(Sect, 'Shift', False);
            if (modControl in Modifiers) then
              WriteBool(Sect, 'Control', True)
            else
              WriteBool(Sect, 'Control', False);
            if (modAlt in Modifiers) then
              WriteBool(Sect, 'Alt', True)
            else
              WriteBool(Sect, 'Alt', False);
            if (modWin in Modifiers) then
              WriteBool(Sect, 'Win', True)
            else
              WriteBool(Sect, 'Win', False);
            WriteBool(Sect, 'Enabled', Enabled);
          end;
        end;
      end;
    finally
      IniFile.Free;
    end;
  end
  else
    raise Exception.Create('IniFileName has no value!');
end;

procedure TAMHotKeys.SaveToRegister;
var
  I: integer;
begin
  if Assigned(FAMReg) then
  begin
    with FAMReg do
    begin
      Active := True;
      try
        Reg.DeleteKey(GetKeyBasis);
        for I := 0 to HotKeys.Count - 1 do
        begin
          if Reg.OpenKey(GetKeyBasis + HotKeys[I].DisplayName, True) then
          begin
            Reg.WriteInteger('VirtKey', Ord(HotKeys[I].VirtKey));
            Reg.WriteBool('Shift', (modShift in HotKeys[I].Modifiers));
            Reg.WriteBool('Control', (modControl in HotKeys[I].Modifiers));
            Reg.WriteBool('Alt', (modAlt in HotKeys[I].Modifiers));
            Reg.WriteBool('Win', (modWin in HotKeys[I].Modifiers));
            Reg.WriteBool('Enabled', HotKeys[I].Enabled);
          end;
        end;
      finally
        Active := False;
      end;
    end;
  end
  else
    raise Exception.Create('There is no AMReg-component assigned to AMReg-rpoperty!');
end;

procedure TAMHotKeys.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FAMReg) then
  begin
    FAMReg := nil;
  end;
end;

procedure TAMHotKeys.Loaded;
var
  I: integer;
begin
  if not (csDesigning in ComponentState) then
    for I := 0 to HotKeys.Count - 1 do
      if HotKeys[I].Enabled then HotKeys[I].RegisterHK;
end;

constructor TAMHotKeys.Create(AOwner: TComponent);
var
  I: integer;
begin
  for I := 0 to AOwner.ComponentCount - 1 do
    if AOwner.Components[I] is TAMHotKeys then
      raise Exception.Create('No more than one TAMHotKeys-component on a form');

  if (AOwner =  nil) or not (AOwner is TForm) then
    raise Exception.Create('The owner of TAMHotKeys must be a form');

  inherited Create(AOwner);
  FHotKeys := THotKeys.Create(Self);
  FSaveLoad := slRegister;

  NewWndProc := nil;
  OldWndProc := nil;
  HookParent;
end;

destructor TAMHotKeys.Destroy;
begin
  FHotKeys.Free;
  UnhookParent;
  inherited Destroy;
end;

// THotKeys implementation
function THotKeys.GetItem(Index: Integer): THotKeyItem;
begin
  Result := THotKeyItem(inherited GetItem(Index));
end;

procedure THotKeys.SetItem(Index: Integer; Value: THotKeyItem);
begin
  inherited SetItem(Index, Value);
end;

function THotKeys.GetOwner: TPersistent;
begin
  Result := FAMHotKeys;
end;

function THotKeys.Add: THotKeyItem;
begin
  Result := THotKeyItem(inherited Add);
end;

constructor THotKeys.Create(AMHotKeys: TAMHotKeys);
begin
  inherited Create(THotKeyItem);
  FAMHotKeys := AMHotKeys;
end;

// THotKeyItem implementation
procedure THotKeyItem.SetEnabled(Value: boolean);
begin
  if (Value <> FEnabled) then
  begin
    FEnabled := Value;

    if not (csDesigning in TAMHotKeys(THotKeys(GetOwner).GetOwner).ComponentState) then
      if FEnabled then
        RegisterHK
      else
        UnregisterHK;
  end;
end;

procedure THotKeyItem.SetModifiers(Value: TModifiers);
begin
  if (Value <> FModifiers) and (VirtKey <> vkNone) then FModifiers := Value;
end;

procedure THotKeyItem.SetVirtKey(Value: TVirtKey);
begin
  if Value <> FVirtKey then
  begin
    FVirtKey := Value;
    if Value = vkNone then
    begin
      Enabled := False;
      Modifiers := [];
    end;
  end;
end;

function THotKeyItem.GetDisplayName: string;
begin
  Result := '';
  if modShift    in Modifiers then Result := Result + 'Shift';
  if modControl  in Modifiers then Result := Result + 'Control';
  if modAlt      in Modifiers then Result := Result + 'Alt';
  if modWin      in Modifiers then Result := Result + 'Win';
  if Result = '' then
    Result := VirtKeyStr[VirtKey]
  else
    Result := Result + '_' + VirtKeyStr[VirtKey];

  if Result = '' then Result := inherited GetDisplayName;
end;

procedure THotKeyItem.Assign(Source: TPersistent);
begin
  if Source is THotKeyItem then
  begin
    Enabled := THotKeyItem(Source).Enabled;
    Modifiers := THotKeyItem(Source).Modifiers;
    VirtKey := THotKeyItem(Source).VirtKey;
    Exit;
  end;
  inherited Assign(Source);
end;

procedure THotKeyItem.RegisterHK;

  function ModToFlag(Modifiers : TModifiers): UInt;
  begin
    Result := 0;
    if modShift    in Modifiers then Result := Result or MOD_SHIFT;
    if modControl  in Modifiers then Result := Result or MOD_CONTROL;
    if modAlt      in Modifiers then Result := Result or MOD_ALT;
    if modWin      in Modifiers then Result := Result or MOD_WIN;
  end;

begin
  FRegistered := False;
  if (VirtKey <> vkNone) then
    FRegistered := RegisterHotKey((TAMHotKeys(THotKeys(GetOwner).GetOwner).Owner as TWinControl).Handle,
                                   ID, ModToFlag(Modifiers), VirtKeys[VirtKey]);
end;

procedure THotKeyItem.UnregisterHK;
begin
  if FRegistered then
    FRegistered := not UnregisterHotKey((TAMHotKeys(THotKeys(GetOwner).GetOwner).Owner as TWinControl).Handle, ID);
end;

procedure THotKeyItem.Free;
begin
  if Registered then Enabled := False;
  inherited Free;
end;

constructor THotKeyItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FEnabled := False;
  FRegistered := False;
  FVirtKey := vkNone;
end;

procedure Register;
begin
  RegisterComponents('AM Software', [TAMHotKeys]);
end;

end.
