unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, AMHotKey, Buttons, ExtCtrls, IniFiles, AMReg;

type
  TForm1 = class(TForm)
    AMHK: TAMHotKeys;
    GroupBox1: TGroupBox;
    cbShift: TCheckBox;
    cbControl: TCheckBox;
    cbAlt: TCheckBox;
    cbWin: TCheckBox;
    GroupBox2: TGroupBox;
    cmVK: TComboBox;
    cbEnabled: TCheckBox;
    GroupBox3: TGroupBox;
    lbHotKeys: TListBox;
    btnAdd: TButton;
    btnDelete: TButton;
    btnModify: TButton;
    cbOnHotKeys: TCheckBox;
    Bevel1: TBevel;
    Bevel2: TBevel;
    btnSave: TButton;
    btnLoad: TButton;
    btnDeleteAll: TButton;
    btnClose: TButton;
    btnUp: TBitBtn;
    btnDown: TBitBtn;
    AMR: TAMReg;
    GroupBox4: TGroupBox;
    rbRegister: TRadioButton;
    rbIniFile: TRadioButton;
    procedure FormShow(Sender: TObject);
    procedure btnAddClick(Sender: TObject);
    procedure cmVKChange(Sender: TObject);
    procedure lbHotKeysClick(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure btnModifyClick(Sender: TObject);
    procedure cbOnHotKeysClick(Sender: TObject);
    procedure btnLoadClick(Sender: TObject);
    procedure btnDeleteAllClick(Sender: TObject);
    procedure Button1Click(Sender: TObject);
    procedure btnUpClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    procedure OnHotKeyPressed(Sender: TObject);
    procedure OnHotKeysPressed(Sender: TObject; HotKeyIndex: Word);
    procedure UpdateBtns;
    procedure UpdateHKList;
    procedure UpdateVKList;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

// General "predefined" OnHotKey-event
procedure TForm1.OnHotKeyPressed(Sender: TObject);
begin
  MessageBeep(0);
  ShowMessage('Hello there is a hot-key combination used!');
end;

// General "predefined" OnHotKeys-event
procedure TForm1.OnHotKeysPressed(Sender: TObject; HotKeyIndex: Word);
begin
  MessageBeep(1);
  ShowMessage(Format('The following hotkey-combination is pressed: %d', [HotKeyIndex]));
end;

// Update Button-state depending on some settings
procedure TForm1.UpdateBtns;
begin
  btnAdd.Enabled := cmVK.ItemIndex <> -1;
  btnDelete.Enabled := lbHotKeys.ItemIndex <> -1;
  btnUp.Enabled := lbHotKeys.ItemIndex <> -1;
  btnDown.Enabled := lbHotKeys.ItemIndex <> -1;
  btnModify.Enabled := (cmVK.ItemIndex <> -1) and (lbHotKeys.ItemIndex <> -1);
end;

// When we load the appl. we must create a list of all available Virtual keys
procedure TForm1.UpdateVKList;
var
  I: integer;
begin
  cmVK.Clear;
  for I := 0 to 102 do
  begin
    cmVK.Items.Add(VirtKeyStr[TVirtKey(I)]);
  end;
end;

// When we change AMHotKeys we must update de list with "Available Hotkey Combinations"
procedure TForm1.UpdateHKList;
var
  I: integer;
begin
  lbHotKeys.Clear;
  for I := 0 to AMHK.HotKeys.Count - 1 do lbHotKeys.Items.Add(AMHK.HotKeys[I].DisplayName);
    
  UpdateBtns;
end;

// Execute two procedures defined above
procedure TForm1.FormShow(Sender: TObject);
begin
  AMHK.IniFileName := ExtractFilePath(ParamStr(0)) + AMHK.IniFileName;
  UpdateVKList;
end;

// Code to add a new hotkey defenition depending on the settings (Virt. key, Modifiers and Enabled)
procedure TForm1.btnAddClick(Sender: TObject);
var
  HKItem : THotKeyItem;
  I, J: Integer;
begin
  try
    HKItem := AMHK.HotKeys.Add;
    HKItem.VirtKey := TVirtKey(cmVK.ItemIndex);
    if cbShift.Checked then HKItem.Modifiers := HKItem.Modifiers + [modShift];
    if cbControl.Checked then HKItem.Modifiers := HKItem.Modifiers + [modControl];
    if cbAlt.Checked then HKItem.Modifiers := HKItem.Modifiers + [modAlt];
    if cbWin.Checked then HKItem.Modifiers := HKItem.Modifiers + [modWin];
    HKItem.OnHotKey := OnHotKeyPressed;
    HKItem.Enabled := cbEnabled.Checked;
  finally
    with AMHK do
    begin
      for I := HotKeys.Count - 1 downto 0 do
        for J := 0 to HotKeys.Count - 1 do
        begin
          if (HotKeys[I].VirtKey = HotKeys[J].VirtKey) and
             (HotKeys[I].Modifiers = HotKeys[J].Modifiers) and
             (I <> J) then
             begin
               HotKeys[J].Free;
               raise Exception.Create('Duplicate HotKey-defenitions are not allowed!'); 
             end;
        end;
    end;
    UpdateHKList;
  end;
end;

// Update buttons
procedure TForm1.cmVKChange(Sender: TObject);
begin
  UpdateBtns;
end;

// Set all components (Modifiers, Virt. key and Enabled) depending on the selected Hotkey-defenition
procedure TForm1.lbHotKeysClick(Sender: TObject);
begin
  UpdateBtns;
  cmVK.ItemIndex := Ord(AMHK.HotKeys[lbHotkeys.ItemIndex].VirtKey);
  cmVK.OnChange(nil);

  with AMHK.HotKeys[lbHotKeys.ItemIndex] do
  begin
    cmVK.ItemIndex := Ord(VirtKey);
    cbShift.Checked := (modShift in Modifiers);
    cbControl.Checked := (modControl in Modifiers);
    cbAlt.Checked := (modAlt in Modifiers);
    cbWin.Checked := (modWin in Modifiers);
    cbEnabled.Checked := Enabled;
  end;
end;

// Delete the selected Hotkey-defenition
procedure TForm1.btnDeleteClick(Sender: TObject);
begin
  AMHK.HotKeys[lbHotKeys.ItemIndex].Free;

  UpdateHKList;
end;

// Change a SELECTED Hotkey defenition according to the new settings
procedure TForm1.btnModifyClick(Sender: TObject);
begin
  with AMHK.HotKeys[lbHotKeys.ItemIndex] do
  begin
    if Enabled then Enabled := False;
    try
      VirtKey := TVirtKey(cmVK.ItemIndex);
      if cbShift.Checked then
        Modifiers := Modifiers + [modShift]
      else
        Modifiers := Modifiers - [modShift];
      if cbControl.Checked then
        Modifiers := Modifiers + [modControl]
      else
        Modifiers := Modifiers - [modControl];
      if cbAlt.Checked then
        Modifiers := Modifiers + [modAlt]
      else
        Modifiers := Modifiers - [modAlt];
      if cbWin.Checked then
        Modifiers := Modifiers + [modWin]
      else
        Modifiers := Modifiers - [modWin];

      OnHotKey := OnHotKeyPressed;
      Enabled := cbEnabled.Checked;
    finally
      UpdateHKList;
      UpdateBtns;
    end;
  end;
end;

// Assign the predefined OnHotKeysPressed-event to the general AMHotKeys.OnHotKeys-event when checked
// or remove when unchecked
procedure TForm1.cbOnHotKeysClick(Sender: TObject);
begin
  if cbOnHotKeys.Checked then
    AMHK.OnHotKeys := OnHotKeysPressed
  else
    AMHK.OnHotKeys := nil;
end;

// Change the Hotkey-order by selecting a Hotkey-defenition and use Up/Down button
procedure TForm1.btnLoadClick(Sender: TObject);
var
  I: integer;
begin
  if rbRegister.Checked then
    AMHK.SaveLoadFrom := slRegister
  else
    AMHK.SaveLoadFrom := slIniFile;
  AMHK.Load;
  for I := 0 to AMHK.HotKeys.Count - 1 do
  begin
    if AMHK.HotKeys[I].Registered then AMHK.HotKeys[I].OnHotKey := OnHotKeyPressed;
  end;
  UpdateHKList;
end;

// Delete all defined Hotkeys-combinations
procedure TForm1.btnDeleteAllClick(Sender: TObject);
begin
  AMHK.HotKeys.Clear;
  UpdateHKList;
end;

// Close the application
procedure TForm1.Button1Click(Sender: TObject);
begin
  Close;
end;

// Move the selected Hotkey-combination up or down in the list
procedure TForm1.btnUpClick(Sender: TObject);
var
  OldIndex, NewIndex: Integer;
  I: Integer;
begin
  OldIndex := lbHotKeys.ItemIndex;
  NewIndex := OldIndex;
  case TButton(Sender).Tag of
    0 : begin              // Move Up
          if lbHotKeys.ItemIndex > 0 then NewIndex := lbHotKeys.ItemIndex - 1;
        end;
    1 : begin              // Move Down
          if lbHotKeys.ItemIndex < lbHotKeys.Items.Count - 1 then NewIndex := lbHotKeys.ItemIndex + 1;
        end;
  end;
  if NewIndex <> OldIndex then
  begin
    AMHK.HotKeys[lbHotKeys.ItemIndex].Index := NewIndex;
    UpdateHKList;
    for I := 0 to AMHK.HotKeys.Count - 1 do
    begin
      if AMHK.HotKeys[I].Registered then
      begin
        AMHK.HotKeys[I].Enabled := False;
        AMHK.HotKeys[I].Enabled := True;
      end;
    end;
    lbHotKeys.ItemIndex := NewIndex;
    UpdateBtns;
  end;
end;

// Save the defined Hotkey-combination using the build-in Save-method
procedure TForm1.btnSaveClick(Sender: TObject);
begin
  if rbRegister.Checked then
    AMHK.SaveLoadFrom := slRegister
  else
    AMHK.SaveLoadFrom := slIniFile;
  AMHK.Save;
end;

// Automatically save HotKey-settings
procedure TForm1.FormDestroy(Sender: TObject);
begin
  if rbRegister.Checked then
    AMHK.SaveLoadFrom := slRegister
  else
    AMHK.SaveLoadFrom := slIniFile;
  AMHK.Save;
end;

end.
